/*
 * Semtech PS/2 and serial touchscreen driver for 2.6 kernel series
 *
 * Written in 2005 by Jean-Christophe Duberga
 */

/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

/*
 * This driver can handle Semtech (was USAR) Screencoder connected throught 
 * a PS/2 as well as a serial (RS232) port.  
 * This driver can handle the 4 bytes packet protocol as described in 
 * in the Semtech Extended PS/2 Protocol Reference.
 * 
 * It is (at least) suitable with the UR7HCTS2 chip (which support PS/2,
 * RS232 as well as USB interface) when connected via a PS/2 or a RS232
 * interface.
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/input.h>
#include <linux/serio.h>
#include <linux/init.h>

#define DRIVER_DESC	"Semtech psaux/serial touchscreen driver"

MODULE_AUTHOR("Jean-Christophe Duberga <jcduberga@gmx.de>");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

/*
 * Definitions & global arrays.
 */

#define SEMTECH_RESET_COMMAND         0xff
#define SEMTECH_SET_SAMPLING_RATE     0xf3
#define SEMTECH_SAMPLING_RATE_80      0x50
#define SEMTECH_SAMPLING_RATE_60      0x3c
#define SEMTECH_SAMPLING_RATE_100     0x64
#define SEMTECH_READ_DEVICE_TYPE      0xf2
#define SEMTECH_ENABLE_COMMAND        0xf4

#define SEMTECH_ACKNOWLEDGE           0xfA
#define SEMTECH_DEVICE_TYPE           0x11
#define SEMTECH_CMD_RESET             0xFF

#define SEMTECH_MAX_PACKET_LENGTH     4

#define SEMTECH_ABS_MASK              0x8
#define SEMTECH_ABS_BUT_MASK          0x07


// From note 3 ( page 10 )
// C3-C0: Class ID  0x7  UR7HC
// F4-F0: Family ID 0x0A TS2
// P2-P0: Reserved
// V8-V0: Revision  0x60 Current Revision = 0x60
  
unsigned char semtech_init_sequence[]={0xf3, 0x50, 0xf3, 0x3c, 0xf3, 0x64, 0xf2};

#define SEMTECH_INIT_SEQUENCE_LENGTH  (sizeof(semtech_init_sequence))

/*
 * Per-touchscreen data.
 */

enum state 
{
   INIT,
   ENABLE,
   SYNC,
   OUT_OF_SYNC,
   DEVICE_NOT_DETECTED
};

enum pressure_state 
{
   RELEASED,
   PRESSED
};

struct semtech {
	struct input_dev *dev;
	struct serio *serio;
	int id;
	int idx;
	unsigned char csum;
	unsigned char data[SEMTECH_MAX_PACKET_LENGTH];
	char phys[32];
        enum state state; 
        unsigned int init_idx;
        enum pressure_state pressure_state;
};



static irqreturn_t semtech_interrupt(struct serio *serio,
		unsigned char data, unsigned int flags)
{
	struct semtech* semtech = serio_get_drvdata(serio);

        switch(semtech->state)
        {
	 case INIT:
	   
	   if ( semtech->init_idx < SEMTECH_INIT_SEQUENCE_LENGTH )
	     {
		/* Check for acknowledge */
		
		if ( data == SEMTECH_ACKNOWLEDGE )
		  {
		     /* Write the next byte */
		     serio_write(serio, semtech_init_sequence[semtech->init_idx++]);
		  }
		else
		  {
		     printk(KERN_WARNING "semtech: bad acknolowedge received 0x%0x\n", data);
		     semtech->state = DEVICE_NOT_DETECTED;
		  }
	     }
	   else if ( semtech->init_idx == SEMTECH_INIT_SEQUENCE_LENGTH )
	     {
		semtech->init_idx++;
	     }
	   else
	     {
		if ( data == SEMTECH_DEVICE_TYPE )
		  {
		     printk(KERN_WARNING "semtech: Semtech UR7HCTS2 detected\n");		  
		     
		     serio_write(serio, SEMTECH_ENABLE_COMMAND);
		     
		     semtech->idx = 0;
		     semtech->pressure_state = RELEASED;
		     semtech->state = ENABLE;
		  }
		else
		  {
		     printk(KERN_WARNING "semtech: Semtech UR7HCTS2 initialization failure\n");		  
	     
		     semtech->state = DEVICE_NOT_DETECTED;  
		  }	
	     }
	   
	   break;
	 
	 case ENABLE:
	   
	   // Against what stand in the documentation read an extra
	   // acknowledge
	   
	   if ( data == SEMTECH_ACKNOWLEDGE )
	     {
		semtech->state = SYNC;
		printk(KERN_WARNING "semtech: UR7HCTS2 successfully initialized in absolute mode\n");
	     }
	   else
	     {
		semtech->state = OUT_OF_SYNC;
		printk(KERN_WARNING "semtech: UR7HCTS2 initialization failure\n");		  
	     }
	   
	   break;
	   
	 case SYNC:
	   
	   // fill the packet
		
	   semtech->data[semtech->idx++] = data; 
			      
	   if ( ( semtech->idx == 4 ) && ( ( semtech->data[0] & SEMTECH_ABS_MASK ) == 0 ) )
	     {
		int x,y,released;
		
		// Sanity check
		     
		// page 4:   An absolute position report where
		//         b0-b2 of byte 1 are neither 000 nor
		//         111 is invalid, and must be ignored
		//         by the driver.
		
		if ( semtech->data[3] 
		     || ( ( (semtech->data[0] & SEMTECH_ABS_BUT_MASK ) != 0x00 )
		          && ( ( (semtech->data[0] & SEMTECH_ABS_BUT_MASK ) != SEMTECH_ABS_BUT_MASK ) ) ) )
		  {
		     printk(KERN_WARNING "semtech: bad packet received [0x%0x,0x%0x,0x%0x,0x%0x]\n",
			    semtech->data[0],
			    semtech->data[1],
			    semtech->data[2],
			    semtech->data[3]);
		     semtech->state = OUT_OF_SYNC;
		     semtech->idx=0;
		     break;
		  }
		
		/* page 7: EXTENDED PS/2 ABSOLUTE POSITION REPORT
		 *
		 * Byte 1
		 *   b0 |                     
		 *   b1 |- Touch           000 = touch, 111 = no touch
		 *   b2 |
		 *   b3    0               Always = 0 (absolute)
		 *   b4    X8
		 *   b5    X9              MSb of X data
		 *   b6    Y8
		 *   b7    Y9              MSb of Y data
		 * Byte 2
		 *   b0    X0              LSb of X data
		 *   b1    X1
		 *   b2    X2
		 *   b3    X3
		 *   b4    X4
		 *   b5    X5
		 *   b6    X6
		 *   b7    X7
		 * Byte 3
		 *   b0    Y0              LSb of Y data
		 *   b1    Y1
		 *   b2    Y2
		 *   b3    Y3
		 *   b4    Y4
		 *   b5    Y5
		 *   b6    Y6
		 *   b7    Y7
		 * Byte 4
		 *  Always NULL
		*/
		
		
		x = ((semtech->data[0] & 0x30) << 4) | semtech->data[1];
		y = ((semtech->data[0] & 0xc0) << 2) | semtech->data[2];
		released = ( semtech->data[0] & SEMTECH_ABS_BUT_MASK );
		
		input_report_abs(semtech->dev, ABS_X, x);
		input_report_abs(semtech->dev, ABS_Y, y);
		
		/* Report a key touch/release only when this state change */ 
		
		if ( semtech->pressure_state == RELEASED && !released )
		  {
		     semtech->pressure_state = PRESSED;
		     input_report_key(semtech->dev, BTN_TOUCH, 1);
		  }
		else if ( semtech->pressure_state == PRESSED && released)
		  {
		     semtech->pressure_state = RELEASED;
		     input_report_key(semtech->dev, BTN_TOUCH, 0);
		  }
		
		input_sync(semtech->dev);
		
		semtech->idx = 0;

		printk(KERN_WARNING "semtech: [ABS] %c %d %d\n",released?' ':'+',x,y );
	     }
	  else if ( semtech->idx == 4 ) /* Relative packet i.e. from external PS/2 mouse */
	     {
		int x,y;
		char *signed_data = (char *)(semtech->data);
		
		x = ((signed_data[0] & 0x10) << 4) | signed_data[1];
		y = ((signed_data[0] & 0x20) << 3) | signed_data[2];
	
		input_report_rel(semtech->dev, REL_X,  x);
		input_report_rel(semtech->dev, REL_Y, -y);
		input_report_key(semtech->dev, BTN_LEFT,   semtech->data[0] & 0x01);
		input_report_key(semtech->dev, BTN_RIGHT,  semtech->data[0] & 0x02);
		input_report_key(semtech->dev, BTN_MIDDLE, semtech->data[0] & 0x04);
	 
		input_sync(semtech->dev);
		
		printk(KERN_WARNING "semtech: [REL] %d %d\n", x, y);	     
		semtech->idx = 0;
	     }
	   
	   break;
	   
	 case OUT_OF_SYNC:
	   
	   printk(KERN_WARNING "semtech: out of sync byte received 0x%0x\n", data);		  
		
	   /* Wait for a null byte, that we hope to be the end of a valid packet */
	   
	   if ( data == 0 )
	     {
		semtech->idx = 0;
		semtech->state = SYNC;
	     }
	   
	   break;
	   
	 case DEVICE_NOT_DETECTED:
	   
	   break;
	}
   
        return IRQ_HANDLED;
}

/*
 * semtech_disconnect() is the opposite of semtech_connect()
 */

static void semtech_disconnect(struct serio *serio)
{
	struct semtech* semtech = serio_get_drvdata(serio);

	input_unregister_device(semtech->dev);
	serio_close(serio);
	serio_set_drvdata(serio, NULL);
	kfree(semtech);
}

/*
 * semtech_connect() is the routine that is called when the semtech input 
 * driver is registred to handle a serio device.
 */

static int semtech_connect(struct serio *serio, struct serio_driver *drv)
{
	struct semtech *semtech;
	struct input_dev *input_dev;
	int err;

	semtech = kzalloc(sizeof(struct semtech), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!semtech || !input_dev) {
		err = -ENOMEM;
		goto fail;
	}

	semtech->serio = serio;
	semtech->id = serio->id.id;
	semtech->dev = input_dev;
        semtech->state = INIT;
        semtech->init_idx = 0;
   
	snprintf(semtech->phys, sizeof(semtech->phys), "%s/input0", serio->phys);

	input_dev->private    = semtech;
	input_dev->name       = "semtech psaux/serial TouchScreen";
	input_dev->phys       = semtech->phys;
	input_dev->id.bustype = BUS_I8042;
	input_dev->id.vendor  = SERIO_SEMTECH;
	input_dev->id.product = semtech->id;
	input_dev->id.version = 0x0100;
	input_dev->cdev.dev   = &serio->dev;

	input_dev->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	input_dev->keybit[LONG(BTN_TOUCH)] = BIT(BTN_TOUCH);
        
        input_set_abs_params(input_dev, ABS_X, 0, 1023, 0, 0);
        input_set_abs_params(input_dev, ABS_Y, 0, 1023, 0, 0);
        input_set_abs_params(input_dev, ABS_PRESSURE, 0, 255, 0, 0);
   
        input_dev->relbit[0] = BIT(REL_X) | BIT(REL_Y);     
        input_dev->evbit[0] |= BIT(EV_REL);
        input_dev->keybit[LONG(BTN_MOUSE)] |= BIT(BTN_LEFT) | BIT(BTN_MIDDLE) | BIT(BTN_RIGHT);

        serio_set_drvdata(serio, semtech);

	err = serio_open(serio, drv);
	if (err)
		goto fail;

        serio_write(serio, semtech_init_sequence[semtech->init_idx++]);
   
	input_register_device(semtech->dev);
	return 0;

 fail:	serio_set_drvdata(serio, NULL);
	input_free_device(input_dev);
	kfree(semtech);
	return err;
}

/*
 * The serio driver structure.
 */

static struct serio_device_id semtech_serio_ids[] = {
	{
		.type	= SERIO_ANY,
//		.proto	= SERIO_SEMTECH,
		.id	= SERIO_ANY,
		.extra	= SERIO_ANY,
	},
	{ 0 }
};

MODULE_DEVICE_TABLE(serio, semtech_serio_ids);

static struct serio_driver semtech_drv = {
	.driver		= {
		.name	= "semtech",
	},
	.description	= DRIVER_DESC,
	.id_table	= semtech_serio_ids,
	.interrupt	= semtech_interrupt,
	.connect	= semtech_connect,
	.disconnect	= semtech_disconnect,
};

/*
 * The functions for inserting/removing us as a module.
 */

static int __init semtech_init(void)
{
	serio_register_driver(&semtech_drv);
	return 0;
}

static void __exit semtech_exit(void)
{
	serio_unregister_driver(&semtech_drv);
}

module_init(semtech_init);
module_exit(semtech_exit);
